# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import unicode_literals

"""
This package contains various command line interfaces to programs used in
pymatgen that do not have Python equivalents.
"""
